/**************************************************\
*                 C S 9 1 2 8 0                    * 
*           (c) 2014 - Daniel Coulom               *  
*           http://dcmoto.free.fr/                 *
*           http://forum.system-cfg.com/           *
*--------------------------------------------------*
* Ce code est distribue gratuitement dans l'espoir *
* qu'il sera utile, mais sans aucune  garantie  et *
* sans  engager  la  responsabilit  de  l'auteur. *
* Vous  pouvez  l' utiliser,  le  modifier  et  le *
* diffuser librement, en conservant cette  licence *
* et les rfrences de l'auteur dans   toutes  les *
* copies. L'exploitation commerciale est interdite.*
\**************************************************/

* Ce programme emule un controleur de disquette
* Thomson accedant a une image au format .sd par
* l'intermediaire des interfaces SDMOTO ou SDMO
* Il est destine a etre charge dans l'EPROM du
* controleur nanoreseau, dans les banques 0 et 1.
* La banque 0 accede a l'interface SDMOTO
* La banque 1 accede a l'interface SDMO
* Auteur : Daniel Coulom, d'aprs la rom CD90-640
* Calcul du LBA avec le code de Samuel Devulder
* Optimisation lecture octet par Samuel Devulder

/**************************************************\
*                Version 2016.09.02                *
*     (reporter la version en fin de programme)    *
\**************************************************/
* Historique
* 2016.09.02 RCONF restaure le bit MOTOR OFF pour SDMO     
* 2016.01.23 positionnement bit carry en cas d'erreur     
* 2015.12.17 ajout d'un commentaire en fin de l'EPROM     
* 2015.12.17 suppression de code inutile     
* 2015.02.13 ajout signature SDMO/SDMOTO en $A6FF/$E6FF     
* 2015.02.13 suppression reset TO9 (incompatible)     
* 2014.12.28 pas de reset au demarrage (pour MO6)     
* 2014.12.27 modifications pour compatibilite MO et TO     
* 2014.12.08 correction pour SD avec partition FAT16     
* 2014.12.07 adaptation recherche BOOT.SD pour carte SDHC     
* 2014.12.06 ajout de la compatibilite FAT16 et FAT32
* 2014.12.05 calcul du LBA du fichier BOOT (par Samuel)
* 2014.12.03 recherche fichier BOOT dans le repertoire
* 2014.11.30 lecture du repertoire de la carte SD
* 2014.05.18 adresse disquette de boot en $200 (fixe)
* 2014.03.01 pas de reset pour une carte deja initialisee
* 2014.03.01 indicateur carte initialisee $55 en CMDXX+5
* 2014.02.27 initialisation mosi au niveau haut
* 2014.02.27 montee du signal d'horloge avant lecture bit
* 2014.02.27 EXCMD=attente octet $FF, EXCMD0=pas d'attente
* 2014.02.27 suppression retries d'initialisation carte
* 2014.02.21 abandon EXCMD si la carte n'est pas prete
* 2014.01.08 lecture systematique du LBA0 a chaque reset
* 2014.01.08 3 retries en cas d erreur initialisation SD
* 2014.01.08 abandon de la sortie de la boucle EXCMD
* 2014.01.06 sortie de la boucle EXCMD si carte non prete
* 2014.01.06 x 16 delai initialisation carte -> 1 seconde
* 2013.12.30 correction erreur de compilation cs91281
* 2013.12.21 correction bug de calcul adresse LB0 pour MO
* 2013.12.07 adresse fichier .sd lue sans buffer en ram
* 2013.12.07 suppression des retries lors du boot
* 2013.12.05 ajout fonctions $04 & $10 (selection densite)
* 2013.12.01 retablissement configuration port joysticks
* 2013.10.30 deroulement boucle ecriture octet dans SWRITE
* 2013.10.30 deroulement boucle lecture octet dans SREAD
* 2013.10.29 gain de 2 cycles dans RBYTE version SDMO
* 2013.10.21 suppression initialisation du CNA en sortie
* 2013.10.20 correction initialisation PIA musique et jeu
* 2013.10.19 indirection vers EXCMD en $A028-$E028 
* 2013.10.18 banque 0 ou 1 choisie par parametre -dBANK 
* 2013.10.17 suppression des retries fonctions standard
* 2013.10.17 retour erreur unite / piste / secteur
* 2013.10.10 formatage rapide pistes 0 et 20 seulement
* 2013.10.08 correction detection SDHC (BLS -> BPL)
* 2013.08.26 SB_LB0 decale de 2 octets (evite ecrasement)
* 2013.08.14 amliorations mineures
* 2013.08.11 procedure de reset pour TO8/TO8D/TO9/TO9+
* 2013.08.10 procedure de reset pour TO8D
* 2013.08.08 zones de travail en debut de pile systeme
* 2013.08.07 copie routine RBYTE dans SREAD evite BSR
* 2013.08.07 copie routine WBYTE dans SWRITE evite BSR
* 2013.08.07 test position switches pour choix SD_LB0
* 2013.08.06 buffer en $6300 pour lecture secteur 1
* 2013.08.06 routine CLOCK envoie 8 x A tops horloge
* 2013.08.06 attente carte prete avant envoi commande
* 2013.08.05 test OK en configuration reelle sur MO6
* 2013.08.05 temporisation entre ecritures secteurs
* 2013.08.05 correction erreur dans RECFI
* 2013.07.28 ecriture secteur et formatage OK dans dcmoto
* 2013.07.25 boot, chargement DOS, lecture secteur OK
* 2013.07.24 ajout du calcul LBA ecrit par sam
* 2013.07.23 debut du projet

*------------------------------------------------------
* VECTEURS EN RAM
*------------------------------------------------------
DK_OPC  EQU $6048   code commande controleur disquette
DK_DRV  EQU $6049   numero du lecteur de disquette
DK_TRK  EQU $604a   n piste (2 octets)
DK_SEC  EQU $604c   n secteur
DK_NUM  EQU $604d   entrelacement
DK_STA  EQU $604e   etat courant controleur disquette
DK_BUF  EQU $604f   adresse buffer secteur (2 octets)
* position tetes    n piste + n secteur (2 octets)
TRACK0  EQU $6051   position tete lecteur 0
TRACK1  EQU $6053   position tete lecteur 1
TRACK2  EQU $6055   position tete lecteur 2
TRACK3  EQU $6057   position tete lecteur 3
ROTAT   EQU $6058   indicateur de rotation du moteur
DKFLG   EQU $6080   indicateur presence controleur disque

* Zones de travail en pile systeme pour le driver de la carte SD
SD_LB0  EQU $608E   adresse du debut du fichier .sd dans la carte SD
SD_TYP  EQU $6092   type de carte SD=0 SDHC=1
CMDXX   EQU $6093   definition de la commande CMD17 ou CMD24
SD_LBA  EQU $6094   adresse du secteur courant dans la carte SD

DKWE5   EQU $60e5   ????
DKWE7   EQU $60e7   pointeur sur nom de fichier
DKWE9   EQU $60e9   pointeur sur buffer
DKWEB   EQU $60eb   type de fichier
DKWEC   EQU $60ec   flag de fichier 
DKWED   EQU $60ed   bloc libre dans la FAT
DKWF0   EQU $60f0   code operation logique
DKWF5   EQU $60f5   numero de secteur
DKWF6   EQU $60f6   numero de bloc  
DKWF7   EQU $60f7   nombre d'octets dans le dernier secteur du fichier
DKWF9   EQU $60f9   numero de bloc alloue
DKWFA   EQU $60fa   numero du premier secteur du bloc
DKWFB   EQU $60fb   numero de piste du bloc courant
BUFFER  EQU $6200   buffer de lecture/ecriture secteur

* Zones de travail pour monter la disquette de demarrage
BUF512  EQU $9C00   adresse buffer de lecture secteur 512 octets
CSIZ2   EQU $9E00   nombre de secteurs par cluster (sur 2 octets)
CSIZE   EQU $9E01   nombre de secteurs par cluster (sur 1 octet)
RSECT   EQU $9E02   nombre de secteurs reserves
PSECT   EQU $9E04   numero secteur debut partition
DSECT   EQU $9E06   numero secteur debut directory
FSECT   EQU $9E08   numero secteur du data cluster 0    

*******************************************************
* DESCRIPTION DES FONCTIONS SPECIALES CS91-280
*
* EXCMD Lance une commande pour la carte SD
* - Registre DP = $A7(MO) ou $E7(TO)
* - Le PIA manettes doit tre initialise pour SDMOTO
* - Registre U = adresse des parametres de la commande
*   (code commande + 4 octets + checksum + code retour)
* - Retour registre A (0 = OK, sinon erreur)
* - Les registres ne sont pas preserves (sauf DP) 
*
* RS512 Lit un secteur de 512 octets
* - CMDXX+1 = LBA secteur,   en $2094(MO) ou $6094(TO)
* - DK_BUF = Adresse buffer, en $204F(MO) ou $604F(TO)
* - Les registres ne sont pas preserves
* - Le registre DP est retabli en $20(MO) ou $60(TO)
*
*******************************************************

*------------------------------------------------------
* ROM DU CONTROLEUR
*------------------------------------------------------
  ORG $E000
  FCB $53        S  controleur carte SD     
  FCB $54        T  fat de 160 octets     
  FCB $44        D  double densite    
  FCB $40        7  checksum 53+54+44+55=140

OPTABL
  LBRA  DKCONT   fonctions standard      
  LBRA  DKBOOT   lancement du boot       
  LBRA  DKFMT    formatage               
  LBRA  LECFA    chargement de la fat    
  LBRA  RECFI    ouverture d'un fichier  
  LBRA  RECUP    effacement d'un fichier 
  LBRA  ECRSE    ecriture d'un secteur   
  LBRA  ALLOD    creation d'un fichier   
  LBRA  ALLOB    allocation d'un bloc    
  LBRA  MAJCL    mise a jour cluster     
  LBRA  FINTR    cloture d'ecriture      
  LBRA  RESETO   reset special TO8/TO8D/TO9/TO9+      
  LBRA  EXCMD    execution commande pour carte SD      
  LBRA  RS512    lecture secteur de 512 octets

*------------------------------------------------------
* Initialisation des registres
*------------------------------------------------------
INIREG
  TFR   PC,D            adresse courante            
  ANDA  #$70            calcul valeur DP
  TFR   A,DP            DP = $20 ou $60
  RTS   
  
*------------------------------------------------------
* DKBOOT = Lancement du boot
*------------------------------------------------------
DKBOOT
  BSR  INIREG           initialisation DP       
  CLR   <DK_DRV         0 dans le numero de lecteur
DKB1
  CLR   <DK_TRK         toujours zero
  CLR   <DK_TRK+1       selection piste 0
  LDA   #$01            valeur pour secteur 1
  STA   <DK_SEC         selection secteur 1
  LBSR  RESET           initialisation carte SD
  BCS   DKB9            erreur d'initialisation
  LEAU  BUFFER,PCR      adresse du buffer secteur
  STU   <DK_BUF         stockage adresse buffer
  LBSR  RS256           lecture secteur logique
  BCS   DKB9            erreur de lecture
  LEAU  BUFFER,PCR      adresse debut de buffer      
  LEAY  BUFFER+$7f,PCR  adresse fin de buffer
  PSHS  Y               empilage
  LDA   #$55            calcul checksum boot
DKB4
  DEC   ,U              octet - 1
  COM   ,U              complement
  ADDA  ,U+             ajout checksum
  CMPU  ,S              test fin de buffer
  BNE   DKB4            octet suivant
  PULS  Y     
  CMPA  ,U              test checksum boot
  BNE   DKB9            checksum fausse
  JMP   BUFFER,PCR      execute le boot
DKB9
  CLR   <DKFLG          Indicateur de presence controleur
  JMP   [$001e]         lancement application a froid

*------------------------------------------------------
* DKCONT = Fonctions standard
*------------------------------------------------------
DKCONT
  PSHS  U,Y,X,DP,B,A,CC                  
  BSR   INIREG          initialisation registres
  CLRA                  status=0
  PSHS  A               empile le status
  PULS  B               depile le status
  BSR   STDOP           execution de l'operation
  PULS  A               depile CC 
  EXG   A,CC            restaure CC
  LSRA                  bit d'erreur dans CC
  PULS  A,B,DP,X,Y,U,PC
  
*------------------------------------------------------
* Execution fonction standard
*------------------------------------------------------
STDOP
  LDA   <DK_OPC        Code operation  
  LEAY  >RESET,PCR     Reset controleur
  BITA  #$01           Bit reset
  BNE   STDOP2         Execution reset          
  LEAY  >RS256,PCR     Lecture secteur logique
  BITA  #$02           Bit lecture secteur
  BNE   STDOP2         Execute lecture secteur   
  LEAY  >SWRITE,PCR    Ecriture secteur logique
  BITA  #$08           Bit ecriture secteur
  BNE   STDOP2         Execute lecture secteur 
  LEAY  >FIND0,PCR     Recherche piste 0
  BITA  #$20           Bit recherche piste 0
  BNE   STDOP2         Execute recherche piste 0
  LEAY  >FINDT,PCR     Recherche piste
  BITA  #$40           Bit recherche piste x
  BNE   STDOP2         Execute recherche piste x
  LEAY  >SIMPLE,PCR    Passage simple densite
  BITA  #$04           Bit passage simple densite
  BNE   STDOP2         Execute passage simple densite
  LEAY  >DOUBLE,PCR    Passage double densite
  BITA  #$10           Bit passage double densite
  BEQ   STDOP8         Erreur fonction inconnue           
STDOP2
  JSR   ,Y             Execution de la fonction
  BCC   STDOP9         Retour sans erreur           
STDOP8
  COMB                 Set Carry = erreur
  RTS
STDOP9
  CLRA                 Clear Carry = pas d'erreur
  RTS        

*------------------------------------------------------
* RESET SPECIAL POUR TO8/TO8D/TO9+
*------------------------------------------------------
RESETO
 IF &BANK=0
  ORCC  #$50         desactive les interruptions 
  LDA   #$60         valeur d'initialisation de DP
  TFR   A,DP         initialisation de DP
  BSR   RESET        initialise la carte SD
  LDS   #$60CC       initialise pointeur pile systeme
  LDB   #$DF         type de disque
  LDA   $FFFF        dernier octet rom
  CMPA  #$EE         test rom TO8 version 1
  BEQ   RESET8       reset TO8 version 1
  CMPA  #$C8         test rom TO8D et TO8 v2
  BEQ   RESETD       reset TO8D et TO8 v2
  CMPA  #$A0         test rom TO9+
  BEQ   RESETP       reset TO9+ 
  JMP   [$FFFE]      sinon reset normal
RESET8
  JMP   $FE50        suite du reset du TO8 version 1
RESETD
  JMP   $FE4C        suite du reset du TO8D et TO8v2
RESETP
  JMP   $FE29        suite du reset du TO9+
 ENDIF
  RTS

*--------------------------------
* Passage en simple densite
*--------------------------------
SIMPLE
  LDA   #$80
  RTS

*--------------------------------
* Passage en double densite
*--------------------------------
DOUBLE
  CLRA  
  RTS   
    
*------------------------------------------------------
* RESET CARTE SD POUR INTERFACE SDMOTO OU SDMO
* Pas de reset a la mise sous tension ($A5 en $207F)
* pour contourner le blocage sur MO6 
* Pas de reset si l'indicateur d'initialisation de
* la carte SD est positionne ($55 en CMDXX+5)
*------------------------------------------------------
RESET
  LDA   <$7F          indicateur initialisation
  CMPA  #$A5          test initialisation
  BEQ   RESET1        pas de reset (pour MO6)
  LDA   <SD_LBA+4     indicateur reset carte SD
  CMPA  #$55          test indicateur de reset
  BNE   POWER         initialisation de la carte
RESET1
  CLRA                clear carry
  RTS 

*------------------------------------------------------
* RESTAURATION CONFIGURATION STANDARD PIA 6821 (SDMOTO)
* RESTAURATION BIT MOTOR OFF (SDMO)
* RESTAURATION VALEUR DE DP ($20 pour MO, $60 pour TO)
*------------------------------------------------------
RCONF
 IF &BANK=0          
  LDA   <$CE          lecture registre de controle A
  ANDA  #$FB          raz bit 2 
  STA   <$CE          selection DDRA
  CLRB                tous les bits a zero  
  STB   <$CC          toutes les lignes en entree
  ORA   #$04          set b2
  STA   <$CE          selection PA
 ENDIF
 IF &BANK=1          
  LDA   #$7F          set bit 3
  STA   <$C2          clock high = MOTOR OFF 
 ENDIF
  TFR   DP,A          valeur actuelle DP
  SUBA  #$87          soustraction de $87
  TFR   A,DP          retablir valeur initiale DP
  RTS

*------------------------------------------------------
* INITIALISATION DP
* CONFIGURATION DU PIA 6821 (SDMOTO)
* OU CONFIGURATION DU PORT MAGNETOPHONE (SDMO)
* POUR ACCES A LA CARTE SD
*------------------------------------------------------
SCONF 
  ORCC  #$50          desactive les interruptions 
  TFR   DP,A          lecture de DP
  ADDA  #$87          ajout de $87
  TFR   A,DP          modifie DP pour acces carte SD
 IF &BANK=0          
  LDA   <$CE          lecture registre de controle A
  ANDA  #$FB          raz bit 2 
  STA   <$CE          selection DDRA
  LDB   #$60          set bits 5 et 6
  STB   <$CC          bits MOSI et CLOCK en sortie
  ORA   #$04          set b2
  STA   <$CE          selection PA
  LDB   #$40          set bit 6
  STB   <$CC          keep the DIN high
 ELSE
  LDB   <$C0          valeur port A PIA systeme
  ORB   #$40          set bit 6
  STB   <$C0          keep the DIN high
 ENDIF
  RTS

*---------------------------------
* Retour en erreur initialisation
*---------------------------------
INITR
  BSR   RCONF         retablissement config PIA 6821 
INITR2
  COMA                1 dans code condition
  RTS                 retour

*------------------------------------------------------
* CMD0 = SOFTWARE RESET
* Precedee de l'envoi de 80 tops d'horloge
* En cas d'erreur abandon de l'initialisation
*------------------------------------------------------
POWER
  BSR   SCONF         configuration PIA 6821
  LDA   #$0A          pour 10 appels de 8 tops
  LBSR  CLOCK         envoi de 10 fois 8 tops horloge 
  LEAU  CMD0,PCR      commande CMD0
  LBSR  EXCMD0        execution commande sans attente
  BCS   INITR         erreur d'initialisation

*------------------------------------------------------
* CMD8 = SEND INTERFACE CONDITION
* Non reconnue par les cartes SD de version < 2.00
* c'est pourquoi le code retour n'est pas teste.
* Lire les 4 derniers octets de la reponse R7
*------------------------------------------------------
  LBSR  EXCMD         execution commande 
  LDA   #$04          nbre d'octets
  LBSR  CLOCK         envoi 4 fois 8 tops horloge 

*------------------------------------------------------
* CMD55 + ACMD41 = INITIALISATION
* Retry en cas de code retour different de zero
* Abandon apres 512 tentatives. 
*------------------------------------------------------
  LDY   #$0200        compteur pour 512 essais
INITSD
  LEAY  -1,Y          decrementation compteur
  BEQ   INITR         erreur apres 512 essais
  LEAU  CMD55,PCR     adresse commande CMD55
  LBSR  EXCMD         execution commande CMD55
  LBSR  EXCMD         execution commande ACMD41
  BCS   INITSD        carte non prete, nouvel essai

*------------------------------------------------------
* CMD58 = LECTURE OCR
* Permet de determiner le type de carte SD ou SDHC
*------------------------------------------------------
GETOCR
  LBSR  EXCMD         execution commande
  BCS   INITR         erreur commande
  CLR   SD_TYP,PCR    carte SD par defaut 
  LBSR  RBYTE         lecture OCR poids fort
  ASLA                isole le type de carte 
  BPL   GETOCR1       carte SD   : SD_TYP=0
  INC   SD_TYP,PCR    carte SDHC : SD_TYP=1
GETOCR1      
  LDA   #$03          pour 3 octets        
  LBSR  CLOCK         lecture OCR (octets 2, 3, 4)
  BSR   RCONF         retablissement config PIA 6821 

*------------------------------------------------------
* Initialisation de la commande CMDXX
* Astuce: la checksum initialisee a $55 sert
* d'indicateur d'initialisation de la carte SD  
*------------------------------------------------------
  LDA   #$55          checksum $55
  STA   <SD_LBA+4     initialise checksum

*------------------------------------------------------
* Lecture du Master Boot Record (1er secteur physique)
* $1FE-$1FF: $55AA (signature)
* Premier descripteur (pour la premiere partition)
* $1BE: boot indicator ($00=non, $80=oui) 
* Si boot indicator different, ce n'est pas un MBR
* $1C6-$1C7: LBA (poids faible en tete) 
*------------------------------------------------------
RMBR
  LDD   #$0000        raz registre D
  STD   <SD_LBA       LBA (2 premiers octets)
  STD   <SD_LBA+2     LBA (2 derniers octets)
  STD   PSECT         raz secteur debut partition
  LDD   #BUF512       adresse du buffer
  STD   <DK_BUF       stockage adresse buffer
  LBSR  RS512         lecture du secteur de 512 octets
  LDD   #$55AA        Signature secteur
  CMPD  $9DFE         Test signature
  BNE   INITR2        Signature incorrecte
  LSL   $9DBE         Boot indicator 
  BNE   ABOOT1        Pas de MBR, donc boot record
  LDA   $9DC7         LBA partition pds fort    
  LDB   $9DC6         LBA partition pds faible  
  STD   PSECT         stockage secteur debut partition
  
*------------------------------------------------------
* Lecture du Boot Record
* Premier secteur au debut de la partition
* Registre D = numero secteur de debut de partition
*------------------------------------------------------
RBOOT
  STD   <SD_LBA+2     stockage pour CMD17 si carte SDHC
  TST   <SD_TYP       test du type de carte
  BNE   RBOOT1        carte SDHC, sinon calcul pour SD
  ASLB                multiplie par 2 (poids faible)  
  ROLA                multiplie par 2 (poids fort)
  STD   <SD_LBA+1     adresse = numero secteur * $200  
  CLRB                raz registre A
  STB   <SD_LBA+3     raz dernier octet adresse
RBOOT1 
  LBSR  RS512         lecture du secteur de 512 octets

*------------------------------------------------------
* Analyse du Boot Sector
* $1FE-$1FF: $55AA (signature)
* $00B-$00C: nombre d'octets par secteur (512)
* $00D: nombre de secteurs par cluster
* $00E-$00F: nombre de secteurs reserves
* $010: nombre de FATs
* $011-$012: nombre entrees directory (FAT16 seulement)
* $016-$017: nombre secteurs FAT (FAT16 seulement)
* $024-$027: nombre secteurs FAT (FAT32 seulement)
* $02C-$02D: numero cluster directory (FAT32)
* Attention: le debut des donnees commence au cluster 2
* - En FAT16: apres la directory
* - En FAT32: apres les FATs
* Pour faciliter les calculs, on retranche 2 fois le
* nombre de secteurs par cluster au debut des donnees.
* 
* Calcul du secteur physique du repertoire principal
* = numero du secteur de debut de partition
* + nombre de secteurs reserves
* + nombre de secteurs FAT * nombre de FAT
*
* Calcul du secteur physique du cluster zero
* = numero du secteur du debut du repertoire principal
* + nombre d'entrees repertoire / 16 (0 en FAT32)
* - nombre de secteurs par cluster * 2
*
*------------------------------------------------------
ABOOT
  LDD   #$55AA        Signature secteur
  CMPD  $9DFE         Test signature
  LBNE  INITR2        Signature incorrecte
ABOOT1
  CLRA                
  LDB   $9C0D         Nombre de secteurs par cluster
  STD   CSIZ2         Stockage nbre secteurs par cluster
  LDA   $9C0F         Blocs reserves pds fort   ex: $00
  LDB   $9C0E         Blocs reserves pds faible ex: $04
  STD   RSECT         stockage nbre secteurs reserves
* taille FAT (en FAT16) dans D  
  LDA   $9C17         Fat16 size pds fort       ex: $00
  LDB   $9C16         Fat16 size pds faible     ex: $F2
  STD   $9C16         stockage taille FAT16
  BNE   ABOOT2        FAT16 si different de zero  
* taille FAT (en FAT32) dans D  
* 2 octets poids fort ignores (simplification abusive ?)
  LDA   $9C25         Fat32 size pds fort       ex: $03
  LDB   $9C24         Fat32 size pds faible     ex: $AB
* taille totale fonction du nombre de FATs
ABOOT2
  DEC   $9C10         Nombre de FAT             ex: $02
  BEQ   ABOOT3        Une seule FAT
  ASLB                FAT size multiplie par 2  ex: $E4 
  ROLA                FAT size multiplie par 2  ex: $01
* Le registre D contient le nombre de secteurs des FATs
ABOOT3   
  ADDD  RSECT         ajout nombre de secteurs reserves
  ADDD  PSECT         ajout secteur de debut de partition
  STD   DSECT         stockage secteur debut directory
  LDA   $9C12         nbre entrees directory (poids fort)
  LDB   $9C11         nbre entrees directory (poids faible)
  LSRA                division par 16
  RORB                pour obtenir
  LSRA                le nombre de secteurs
  RORB                de la directory
  LSRA                En FAT32 le nombre d'entrees  
  RORB                de la directory est nul
  LSRA                Le debut des donnees (cluster 2)
  RORB                commence apres les FATs 
  ADDD  DSECT         ajout secteur debut directory
  SUBD  CSIZ2         retrancher le nbre secteurs/cluster 
  SUBD  CSIZ2         retrancher le nbre secteurs/cluster 
  STD   FSECT         stockage numero secteur du cluster 0
  LDD   DSECT         secteur de debut de directory

*------------------------------------------------------
* LECTURE DE LA DIRECTORY
* D contient le numero du secteur a lire
*------------------------------------------------------
RDSECT
  STD   <SD_LBA+2     stockage pour la commande 17
  TST   <SD_TYP       type de carte 0=SD 1=SDHC
  BNE   RDSEC1        carte SDHC
  ASLB                LBA multiplie par 2 
  ROLA                LBA multiplie par 2
  STD   <SD_LBA+1     LBA multiplie par $100
  CLRB
  STB   <SD_LBA+3     octet de poids faible a zero
RDSEC1
  LBSR  RS512         Lecture du secteur 
  LDX   #BUF512       Adresse du buffer

*------------------------------------------------------
* RECHERCHE FICHIER BOOT DANS LA DIRECTORY
* Le registre X pointe sur l'entree en cours 
*------------------------------------------------------
RDIR1  
  CLRB                zero
  CMPB  ,X            Test premier octet
  LBEQ  INITR2        Fin de repertoire
  LDB   #$42          'B'
  CMPB  ,X            Test premier octet
  BNE   RDIR2
  LDB   #$4F          'O'
  CMPB  1,X           Test deuxieme octet
  BNE   RDIR2
  CMPB  2,X           Test troisieme octet
  BNE   RDIR2
  LDB   #$54          'T'
  CMPB  3,X           Test quatrieme octet
  BNE   RDIR2
  LDB   #$20          ' '
  CMPB  4,X           Test cinquieme octet
  BEQ   FBOOT
RDIR2
  LEAX  32,X          Adresse entree suivante
  CMPX  #$9E00        Fin du buffer
  BNE   RDIR1         Traitement entree suivante
* lecture secteur suivant
  LDD   DSECT         secteur debut directory 
  ADDD  #$0001        secteur suivant  
  STD   DSECT         nouveau debut directory 
  BRA   RDSECT        lecture secteur directory

*------------------------------------------------------
* INITIALISATION DU LBA DE LA DISQUETTE DE DEMARRAGE
* CSIZE = nombre de secteurs par cluster (1 octet)
* FSECT = numero de secteur du cluster 0 (2 octets)
* 26,X = n cluster relatif au boot sector (poids faible)
* 27,X = n cluster relatif au boot sector (poids fort)
*
* LBA0      =    XX YY
* N*CLUSTER = UU VV 00 (UU VV = N*A)
*           + 00 ZZ TT (ZZ TT = N*B)
* LBA       = (UU+carry2) (VV+XX+ZZ+carry1) (TT+YY)
*                     /|\    |          /|\    |
*                      |_____|           |_____|
*
* Code original de Samuel Devulder
*
*LBA0      =    XX YY
*N*CLUSTER = UU VV 00 (UU VV = N*A)
*          + 00 ZZ TT (ZZ TT = N*B)
*LBA       = (UU+carry2) (VV+XX+ZZ+carry1) (TT+YY)
*                    /|\    |          /|\    |
*                     |_____|           |_____|
*
*        STB  ,-S        ; sauve lsb(CLUSTER)
*        LDB  N
*        MUL             ; D = msb(CLUSTER)*N = UU VV
*        ADDB SD_LB0+2   
*        ADCA #0         ; D = UU (VV+XX). Nota: Si SD_LB0+1 contient dj 0, ADDD SD_LB0+1 suffit.
*        STD  SD_LB0+1   ; sauv en SD_LB0+1
*        LDB  N
*        LDA  ,S+
*        MUL             ; D = lsb(CLUSTER)*N = ZZ TT
*        ADDD SD_LB0+2   ; D = (VV+XX+ZZ+carry1) (TT+YY)
*        STD  SD_LB0+2   ; (VV+XX+ZZ+carry1) (TT+YY) plac en LB0+2
*        LDB  SD_LB0+1   ; B=UU
*        ADCB #0         ; B=UU+carry2
*        CLRA            ; A=0
*        STD  SD_LB0     ; SD_LB0 = 00 (UU+carry2) (VV+XX+ZZ+carry1) (TT+YY)
*        TST  SD_TYP
*        BNE   FIN       ; ben oui  tout est dj fait
*    ; recopie + decalage
*        STB  ,-S      ; sauvegarde UU+carry2
*        LDD  SD_LB0+2 ; D=(VV+XX+ZZ+carry1) (TT+YY)
*        LSLB
*        ROLA          ; D*2
*        STD  SD_LB0+1 ; sauv en SD_LB0+1
*        LDB  ,S+      ; UU+carry2
*        ROLB          ; (UU+carry2)*2 + "carry de D*2"
*        STB  SD_LB0   ; sauv en SD_B0
*        CLR  SD_LB0+3 ; 0 au bon endroit
*    FIN:
*------------------------------------------------------
FBOOT  
  LDD   FSECT      ; secteur origine
  STD   <SD_LB0+2  ; stockage secteur origine
  LDA   27,X       ; Cluster poids fort
  LDB   CSIZE      ; Nombre de secteurs par cluster
  MUL              ; D = msb(CLUSTER)*N = UU VV
  ADDB  <SD_LB0+2  ; octet de poids fort du secteur origine 
  ADCA  #0         ; D = UU (VV+XX). Nota: Si SD_LB0+1 contient dj 0, ADDD SD_LB0+1 suffit.
  STD   <SD_LB0+1  ; sauv en SD_LB0+1
  LDA   26,X       ; Cluster poids faible
  LDB   CSIZE      ; Nombre de secteurs par cluster
  MUL              ; D = lsb(CLUSTER)*N = ZZ TT
  ADDD  <SD_LB0+2  ; D = (VV+XX+ZZ+carry1) (TT+YY)
  STD   <SD_LB0+2  ; (VV+XX+ZZ+carry1) (TT+YY) plac en LB0+2
  LDB   <SD_LB0+1  ; B=UU
  ADCB  #0         ; B=UU+carry2
  CLRA             ; A=0
  STD   <SD_LB0    ; SD_LB0 = 00 (UU+carry2) (VV+XX+ZZ+carry1) (TT+YY)
  TST   <SD_TYP
  BNE   FBOOT9     ; ben oui  tout est dj fait pour les cartes SDHC
* recopie + decalage pour carte SD
  STB   ,-S        ; sauvegarde UU+carry2
  LDD   <SD_LB0+2  ; D=(VV+XX+ZZ+carry1) (TT+YY)
  LSLB
  ROLA             ; D*2
  STD   <SD_LB0+1  ; sauv en SD_LB0+1
  LDB   ,S+        ; UU+carry2
  ROLB             ; (UU+carry2)*2 + "carry de D*2"
  STB   <SD_LB0    ; sauv en SD_B0
  CLR   <SD_LB0+3  ; 0 au bon endroit
FBOOT9  
  CLRA                0 dans code condition
  RTS                 retour

*------------------------------------------------------
* ATTENTE CARTE PRETE PUIS EXECUTION D'UNE COMMANDE
*------------------------------------------------------
EXCMD
  BSR   RBYTE         lecture d'un octet
  INCA                ajout de 1 ($FF --> $00) 
  BNE   EXCMD         attente carte prete

*------------------------------------------------------
* EXECUTION D'UNE COMMANDE POUR LA CARTE SD
* Le registre B n'est pas preserve.
* le code retour est dans le registre A
*------------------------------------------------------
EXCMD0
  LDX   #$0006        nombre d'octets de commande
EXCMD2
  LDA   ,U+           chargement octet de commande
  BSR   WBYTE         ecriture de l'octet
  LEAX  -1,X          decrementation compteur
  BNE   EXCMD2        il reste des octets a envoyer
EXCMD3
  BSR   ALIGN         lecture d'un octet aligne  
  CMPA  ,U+           test code de retour
  BEQ   EXCMD4        code bon
  COMA                carry set en erreur
  RTS                 retour
EXCMD4
  CLRA                clear carry
  RTS                 retour

 IF &BANK=0
*------------------------------------------------------
* ECRITURE D'UN OCTET POUR SDMOTO
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en entree
*------------------------------------------------------
WBYTE
  LDB   #$08          compteur de boucles = 8
  PSHS  B             empilage compteur de boucles
WBYTE1
  ASLA                b7 reg A -> carry
  RORB                carry -> b7 reg B
  LSRB                b7 reg B -> b6 reg B
  STB   <$CC          envoi bit vers DI
  ORB   #$20          set bit 5
  STB   <$CC          clock high
  ANDB  #$DF          clear bit 5
  STB   <$CC          clock low
  DEC   ,S            decremente compteur boucles
  BNE   WBYTE1        nouvelle boucle
  ORB   #$40          set bit 6
  STB   <$CC          keep the DIN high
  PULS  B,PC          retour

*------------------------------------------------------
* LECTURE D'UN OCTET AVEC ALIGNEMENT SUR BIT 0 (SDMOTO)
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en sortie
* Optimisation transfert b7 avec CMPB par Daniel
*------------------------------------------------------
ALIGN
  CLRA                compteur pour 256 boucles
ALIGN1
  LDB   #$7F          Valeur pour test bit 7
  STB   <$CC          clock high, di high 
  CMPB  <$CC          PA b7 (bit lu) -> carry 
  LDB   #$5F          clear bit 5
  STB   <$CC          clock low, di high
  BCS   ALIGN2        si bit 1, continuer a lire
  LDA   #$FC          sinon stocker bit 0
  BRA   RBYTE1        lire les 7 autres bits
ALIGN2
  DECA                decrementer compteur
  BNE   ALIGN1        nouvelle lecture

*------------------------------------------------------
* LECTURE D'UN OCTET (SDMOTO) = 175 cycles
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en sortie
* Optimisation du compteur de boucle par Samuel
* Optimisation transfert b7 avec CMPB par Daniel
*------------------------------------------------------
RBYTE
  LDA   #$FE          b0 marqueur fin de boucle (2)
RBYTE1      
  LDB   #$7F          Valeur pour test bit 7    (2)
  STB   <$CC          clock high, di high       (4)
  CMPB  <$CC          PA b7 (bit lu) -> carry   (4)
  LDB   #$5F          clear bit 5               (2)
  STB   <$CC          clock low, di high        (4)
  ROLA                C (bit lu) -> b0 reg A    (2)
  BCS   RBYTE1        suite de la boucle        (3) 
  RTS                 retour (octet dans A)     (5)

 ELSE
*------------------------------------------------------
* ECRITURE D'UN OCTET POUR SDMO
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en entree
*------------------------------------------------------
WBYTE
  LDB   #$08          compteur de boucles = 8
  PSHS  B             empilage compteur de boucles
WBYTE1
  LDB   <$C0          lecture port A
  ORB   #$40          b6=1 set DIN
  DEC   ,S            decremente compteur boucles
  BMI   WBYTE3        fin de la boucle
  ASLA                store bit to C
  BCS   WBYTE2        bit=1
  ANDB  #$BF          b6=0 clear DIN
WBYTE2
  STB   <$C0          DIN
  LDB   #$3E          set bit 3
  STB   <$C2          clock high 
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  BRA   WBYTE1        nouvelle boucle
WBYTE3
  STB   <$C0          keep the DIN high
  PULS  B,PC          retour

*------------------------------------------------------
* LECTURE D'UN OCTET AVEC ALIGNEMENT SUR BIT 0 (SDMO)
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en sortie
* Optimisation transfert b7 avec CMPB par Daniel
*------------------------------------------------------
ALIGN
  CLRA                compteur pour 256 boucles
ALIGN1
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          read bit (PA bit 7) dans CC
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  BCS   ALIGN2        si bit 1, continuer a lire
  LDA   #$FC          stocker le bit 0
  BRA   RBYTE1        lire les 7 autres bits
ALIGN2       
  DECA                decrementer compteur
  BNE   ALIGN1        nouvelle lecture

*------------------------------------------------------
* LECTURE D'UN OCTET (SDMO) = 175 cycles
* Le registre B n'est pas prserv
* Valeur de l'octet dans le registre A en sortie
* Optimisation du compteur de boucle par Samuel
* Optimisation transfert b7 avec CMPB par Daniel
*------------------------------------------------------
RBYTE
  LDA   #$FE          b0 marqueur fin de boucle (2)
RBYTE1      
  LDB   #$7F          set bit 3                 (2)  
  STB   <$C2          clock high                (4)
  CMPB  <$C0          read bit (PA b7) dans CC  (4) 
  LDB   #$36          clear bit 3               (2)
  STB   <$C2          clock low                 (4)
  ROLA                C (bit lu) -> b0 reg A    (2)
  BCS   RBYTE1        suite de la boucle        (3) 
  RTS                 retour (octet dans A)     (5)
 ENDIF

*------------------------------------------------------
* Calcul du LBA carte SD : SD_LBA =
* SD_LBA0+(512-511*SD_TYP)*(1280*DK_DRV+16*DK_TRK+DK_SEC-1)
* DK_DRV  n lecteur de disquette
* DK_TRK  n piste (2 octets)
* DK_SEC  n secteur
* SD_TYP  type SD/SDHC
* Routine crite par Samuel
*------------------------------------------------------
SETLBA
  PSHS  U,X,B,A
  LEAU  SD_LBA,PCR
  LDB   <DK_TRK+1  numero de piste
  BMI   SETLB8     erreur secteur
  CMPB  #$50       test piste 80
  BHS   SETLB8     erreur secteur
  LDA   #16
  MUL           ; voir avec lslb rola si les cycles sont importants
  TFR   D,X     ; x = DK_TRK*16
  LDB   <DK_SEC
  BEQ   SETLB8     erreur secteur
  CMPB  #$10       test secteur 16
  BHI   SETLB8     erreur secteur
  DECB
  ABX          ; x = DK_TRK*16 + DK_SEC - 1
  CLRA         ; plus lent que ldd #0, mais 1 octet de moins
  CLRB         
  STD  ,U
  LDA  <DK_DRV
  CMPA  #$04       test unite 4
  BHS   SETLB7     erreur d'unite
  LSLA
  LSLA
  ADDA <DK_DRV
  LEAX D,X     ; x = 1280 * DK_DRV + 16 * DK_TRK + DK_SEC - 1
  TFR  X,D
; la multiplication par 512-511*SD_TYP
  LEAX 4,U
  TST  <SD_TYP
  BNE  SETLB2
  LSLB        
  ROLA        
  CLR  ,-X    
SETLB2
  STD  -2,X   ; <--+  u pointe sur SD_LBA = 0 XX XX 0 ou 0 0 XX XX
; SD_LBA = SD_LBA + SD_LB0 sur 32bits
  LDD  <SD_LB0+2
  ADDD 2,U      
  STD  2,U
  LDD  <SD_LB0
  ADCB 1,U
  ADCA #0    ; en vrai ",u" mais c'est quivalent car on a toujours 0  cette adresse
  STD  ,U
  CLRA              raz bit carry
  BRA   SETLB9
SETLB7  
  LBSR  ERR10       erreur unite  
  BRA   SETLB9
SETLB8  
  LBSR  ERR04       erreur de secteur  
SETLB9
  PULS  A,B,X,U,PC  retour en erreur
  
*------------------------------------------------------
* Lancement de la commande CMD17
* et attente d'un debut de secteur
*------------------------------------------------------
LCMD17
  LDA   #$51       code CMD17
  STA   <CMDXX     initialise le code commande  
  LDY   <DK_BUF    adresse du buffer
  LBSR  SCONF      configuration pour acces carte SD 
  LEAU  CMDXX,PCR  adresse CMD17
  LBSR  EXCMD      execution CMD17 
LCMD18
  BSR   RBYTE      lecture d'un octet
  CMPA  #$FE       comparaison avec $FE
  BNE   LCMD18     attente debut secteur
  RTS              retour
  
*------------------------------------------------------
* Lecture d'un secteur de 512 octets
* LBA pre-initialise dans CMDXX
* Adresse du buffer dans DK_BUF
* Ecriture du buffer indexee par le registre Y
*------------------------------------------------------
RS512
  LBSR  INIREG     initialisation DP ($20 ou $60)
  BSR   LCMD17     lancement de la commande CMD17
* lire 512 octets
  BSR   R512       lecture 512 octets 
* ignorer 2 octets de CRC, retablissement
* de la configuration et retour
  BRA   RS3
  
*------------------------------------------------------
* Lecture d'un secteur de 256 octets
* LBA fonction de: unite, piste, secteur, type carte
* Adresse du buffer dans DK_BUF
* Ecriture du buffer indexee par le registre Y
*------------------------------------------------------
RS256
  BSR   SETLBA     calcul SD_LBA
  LBCS  RS9        erreur unite/piste/secteur 
  BSR   LCMD17     lancement de la commande CMD17
* lire 256 octets
  BSR   R256       lecture 256 octets 
* ignorer 256 octets
  BSR  CLOCK       envoi de 256 * 8 tops
RS3
* ignorer 2 octets de CRC
  LDA   #$02       pour 2 octets
  BSR   CLOCK      ignorer 2 octets
  LBSR  RCONF      retablissement configuration 
RS9
  RTS              retour 
  
*------------------------------------------------------
* Lecture de 512 ou 256 octets d'un secteur
* Le buffer est indexe par le registre Y
*------------------------------------------------------
R512
  BSR   R256
R256
  CLRA             compteur pour 256 boucles 
  PSHS  A          empilage du compteur
R257

 IF &BANK=0
* lecture d'un octet dans A (SDMOTO) *******************
* premier bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* deuxieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* troisieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* quatrieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* cinquieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* sixieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* septieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
* huitieme bit
  LDB   #$7F       Valeur pour test bit 7    (2)
  STB   <$CC       clock high, di high       (4)
  CMPB  <$CC       PA b7 (bit lu) -> carry   (4)
  LDB   #$5F       clear bit 5               (2)
  STB   <$CC       clock low, di high        (4)
  ROLA             C (bit lu) -> b0 reg A    (2)
********************************************************
 ELSE
* lecture d'un octet dans A (SDMO) *********************
* premier bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* deuxieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* troisieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* quatrieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* cinquieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* sixieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* septieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
* huitieme bit
  LDB   #$7F          set bit 3
  STB   <$C2          clock high 
  CMPB  <$C0          PA b7 (bit lu) -> carry
  LDB   #$36          clear bit 3
  STB   <$C2          clock low
  ROLA                C (bit lu) -> b0 reg A    (2)
********************************************************
 ENDIF

  STA   ,Y+        stockage dans le buffer
  DEC   ,S         decrementation compteur 
  BNE   R257       nouvelle lecture
  PULS  A,PC       depilage compteur et retour 

 IF &BANK=0
*------------------------------------------------------
* ENVOI DE TOPS HORLOGE (8 fois reg A) (SDMOTO)
*------------------------------------------------------
CLOCK
  PSHS  A
  LDD   #$7F5F
CLOCK1
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  STA   <$CC          clock high, di high
  STB   <$CC          clock low, di high 
  DEC   ,S
  BNE   CLOCK1                            
  PULS  A,PC                                      
 ELSE
*------------------------------------------------------
* ENVOI DE TOPS HORLOGE (8 fois reg A) (SDMO)
*------------------------------------------------------
CLOCK
  PSHS  A
  LDD   #$3E36
CLOCK1
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  STA   <$C2          clock high, di high
  STB   <$C2          clock low, di high 
  DEC   ,S
  BNE   CLOCK1                            
  PULS  A,PC                                      
 ENDIF
    
*------------------------------------------------------
* Ecriture d'un secteur disquette
* dans secteur SD complete a 512
*------------------------------------------------------
SWRITE
  LBSR  WPROT      test protection ecriture             
  LBLO  SWRIT9     branchement si protection
  LBSR  SETLBA     calcul SD_LBA
  LBCS  SWRIT9     erreur unite/piste/secteur
  LDA   #$58       code CMD24
  STA   <CMDXX     initialise code commande  
  LDY   <DK_BUF    adresse du buffer
  LBSR  SCONF      configuration pour acces carte SD 
  LEAU  CMDXX,PCR  adresse CMD24
  LBSR  EXCMD      execution CMD24 
  LDA   #$FE       indicateur debut bloc 
  LBSR  WBYTE      ecriture de l'octet
* ecrire 256 octets du buffer
  CLRA             compteur pour 256 boucles 
  PSHS  A          empilage du compteur octets
SWRIT2
  LDA   ,Y+        lecture dans le buffer

 IF &BANK=0
* ecriture de l'octet A pour SDMOTO ***************
* premier bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* deuxieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* troisieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* quatrieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* cinquieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* sixieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* septieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* huitieme bit
  ASLA             b7 reg A -> carry
  RORB             carry -> b7 reg B
  LSRB             b7 reg B -> b6 reg B
  STB   <$CC       envoi bit vers DI
  ORB   #$20       set bit 5
  STB   <$CC       clock high
  ANDB  #$DF       clear bit 5
  STB   <$CC       clock low
* set DIN high
  ORB   #$40       set bit 6
  STB   <$CC       keep the DIN high
* test fin de secteur
  DEC   ,S         decrementation compteur octets
  BNE   SWRIT2     nouvelle ecriture
***************************************************
 ELSE
* ecriture de l'octet A pour SDMO *****************
* premier octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* deuxieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* troisieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* quatrieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* cinquieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* sixieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* septieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* huitieme octet
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  ASLA             store bit to C
  BCS   *+4        bit=1
  ANDB  #$BF       b6=0 clear DIN
  STB   <$C0       DIN
  LDB   #$3E       set bit 3
  STB   <$C2       clock high 
  LDB   #$36       clear bit 3
  STB   <$C2       clock low
* set DIN high
  LDB   <$C0       lecture port A
  ORB   #$40       b6=1 set DIN
  STB   <$C0       keep the DIN high
* test fin de secteur
  DEC   ,S         decrementation compteur octets
  LBNE  SWRIT2     nouvelle ecriture
***************************************************
 ENDIF
  
  PULS  A          depilage compteur octets 
* ecrire 256 octets $FF
  LBSR  CLOCK      ecrire 256 octets $FF
* ecrire 2 octets de CRC $FF
  LDA   #$02       pour 2 octets
  LBSR  CLOCK      envoi 2 fois 8 tops horloge
  LBSR  RCONF      retablissement configuration 
SWRIT9
  RTS   

*------------------------------------------------------
* Retour d'un code erreur
*------------------------------------------------------
ERRSET
  STA   <DK_STA
  COMA  
  RTS   

*------------------------------------------------------
* Erreur de secteur
*------------------------------------------------------
ERR04
  LDA   #$04
  BRA   ERRSET

*------------------------------------------------------
* Erreur de piste
*------------------------------------------------------
*ERR02
*  LDA   #$02
*  BRA   ERRSET

*------------------------------------------------------
* Erreur sur les donnees
*------------------------------------------------------
*ERR08
*  LDA   #$08
*  BRA   ERRSET

*------------------------------------------------------
* Erreur unite
*------------------------------------------------------
ERR10
  LDA   #$10
  BRA   ERRSET

*------------------------------------------------------
* Teste la protection ecriture
* Recherche piste
* Test presence disquette
* Recherche piste 0
* Selection densite du logiciel
* Sortie sans erreur
*------------------------------------------------------
WPROT
FINDT
TSTDSK
FIND0
FMTVT
RETOK
  CLRA
INIDEN
  RTS                    retour

*------------------------------------------------------
* DKFMT = formatage de la disquette
*------------------------------------------------------
DKFMT
  PSHS  U,Y,X,DP,B,A,CC              
  ORCC  #$50          desactive les interruptions 
  LBSR  INIREG
*  LBSR  INIOP
*  BSR   INIDEN
*  LBSR  WPROT
*  BLO   DKFMT9  
*  BSR   FIND0      Recherche piste 0
*  BLO   DKFMT9 
  CLR   <DK_TRK       
  CLR   <DK_TRK+1       
DKFMT1
  LBSR  FMTRK      formatage d'une piste 
*  LBSR  FMTVT      verification piste formatee
*  BHS   DKFMT9     erreur de verification

* LDA   <DK_TRK+1  numero de piste
* CMPA  #$4F       derniere piste = 79
* BEQ   DKFMT4     fin de disquette
* INCA             incrementation n piste
* STA   <DK_TRK+1  stockage n piste
* BRA   DKFMT1     piste suivante
DKFMT4
  LBSR  INIT20     initialisation piste 20 
  BLO   DKFMT9 
* LDA   #$28       piste 40   pourquoi ???    
* STA   <DK_TRK+1  stockage n piste          
DKFMT9
  PULS  A 
  EXG   A,CC
  LSRA  
  PULS  A,B,DP,X,Y,U,PC

*------------------------------------------------------
* Initialisation pour operation
*------------------------------------------------------
*INIOP
*  LDA   <DK_OPC
*  ANDA  #$80
*  STA   <DK_OPC
*  CLR   <DK_STA
*  RTS   
  
*------------------------------------------------------
* Formatage d'une piste
*------------------------------------------------------
FMTRK
  LDA   #$E5       caractere de remplissage
  BSR   INIBUF     remplissage buffer
  LDA   #$01       numero premier secteur
  STA   <DK_SEC    stockage numero secteur         
FMTRK1
  LBSR  SWRITE     ecriture d'un secteur
  LDA   <DK_SEC    numero de secteur
  CMPA  #$10       secteur 16
  BEQ   FMTRK9     fin de piste
  INCA             incrementation secteur
  STA   <DK_SEC    stockage numero secteur
  BRA   FMTRK1     secteur suivant
FMTRK9
  RTS              retour
  
*------------------------------------------------------
* Remplissage du buffer de secteur
*------------------------------------------------------
INIBUF
  LDY   <DK_BUF    initialisation adresse
  CLRB             nombre octets secteur
INIB1
  STA   ,Y+        stockage octet remplissage
  DECB             decrementation compteur
  BNE   INIB1      octet suivant 
  RTS              retour
  
*------------------------------------------------------
* Initialisation piste 20
*------------------------------------------------------
INIT20
  LDX   #$0014     piste 20 
  STX   <DK_TRK
  LDA   #$FF       valeur de remplissage
  BSR   INIBUF     remplissage buffer
  LDA   #$08
  ORA   <DK_OPC
  STA   <DK_OPC
  CLRB             initialisation n secteur
INIT21
  INCB             secteur suivant
  STB   <DK_SEC
  LBSR  DKCONT     ecriture du secteur
  BLO   INIT29
  CMPB  #$10       secteur 16
  BNE   INIT21
  LDX   <DK_BUF
  CLR   ,X         premier octet FAT = 0
  LDD   #$FEFE     blocs reserves
  STD   $29,X      pour piste 20
  LDA   #$02
  STA   <DK_SEC
  LBSR  DKCONT     ecriture du secteur
  BLO   INIT29     retour en erreur
  CLRA             pas d'erreur
INIT29
  RTS   

*------------------------------------------------------
* SCRATCH DOS
*------------------------------------------------------
SCRDOS
  FCC   "SCRATCH"
  FCC   " DOS"

*------------------------------------------------------
* FINTR = cloture d'ecriture
*------------------------------------------------------
FINTR
  LDB   <DKWF0     code operation logique   
  CMPB  #$02
  BEQ   FINTR3 
  DEC   <DKWF0     code operation logique
  BSR   RECFI      ouverture fichier
  BLO   FINTR9
  TSTB  
  BEQ   FINTR1
  LBSR  RECUP
  BLO   FINTR9
FINTR1
  INC   <DKWF0     code operation logique
  BSR   RECFI
  BLO   FINTR9
  LDB   #$0a
  LDX   <DKWE7
FINTR2
  LDA   B,X
  STA   B,Y
  DECB  
  BGE   FINTR2
  BSR   ECRSE      ecriture secteur
  BLO   FINTR9
****************** sauvegarde de la FAT
FINTR3
  LDA   #$02
  STA   <DK_SEC    secteur 2
  LDB   #$14
  CLRA  
  STD   <DK_TRK    piste 20
  LDD   <DKWED     adresse de la FAT
  STD   <DK_BUF    adresse du buffer secteur
  BSR   ECRSE      ecriture secteur
  BLO   FINTR9     erreur ecriture
  CLR   <$f0       indicateur cloture fichier
FINTR9
  RTS   

*--------------------------------
* LECFA = Lecture de la FAT
*------------------------------------------------------
LECFA
  LDX   <DKWED     adresse de la FAT
  STX   <DK_BUF    adresse du buffer secteur
  LDA   #$02       numero secteur a lire
  BRA   LSEC20     lecture secteur piste 20

*------------------------------------------------------
* Sortie en erreur
*------------------------------------------------------
RETERR
  STA   <DKWE5
  COMA  
  COMA  
  RTS   

*------------------------------------------------------
* Chargement 1er secteur catalogue
*------------------------------------------------------
LDIR0
  LDA   #$03
  LDX   <DKWE9     pointeur sur buffer
  STX   <DK_BUF

*------------------------------------------------------
* Chargement secteur piste 20
*------------------------------------------------------
LSEC20
  STA   <DK_SEC
  LDB   #$14
  CLRA  
  STD   <DK_TRK
  LDA   #$02
  BRA   SECTIO

*------------------------------------------------------
* ECRSE = Ecriture d'un secteur
*------------------------------------------------------
ECRSE
  LDA   #$08

*------------------------------------------------------
* Operation sur secteur
*------------------------------------------------------
SECTIO
  STA   <DK_OPC
  LDY   <DKWE9     pointeur sur buffer
  LBSR  OPTABL
  LDA   #$03
  RTS   

*------------------------------------------------------
* RECFI = Ouverture d'un fichier
*------------------------------------------------------
RECFI
  BSR   LDIR0      charge 1er secteur catalogue 
RECFI1
  BLO   RETERR     sort si erreur
  LDX   #$0008     compteur de noms dans X
  LDY   <DKWE9     pointeur sur buffer
RECFI2
  LDU   <DKWE7
  LDB   <DKWF0     code operation logique
  CMPB  #$03
  BNE   RECFI3
  LEAU  SCRDOS,PCR adresse du nom de fichier
RECFI3
  CLRB  
RECFI4
  CMPB  #$0b
  BHS   RECFI8
  LDA   B,Y
  CMPA  #$ff
  BEQ   RECFI7
  INCB  
  CMPA  ,U+
  BEQ   RECFI4
  LEAY  $20,Y
  LEAX  -$01,X
  BNE   RECFI2
  INC   <DK_SEC
  LDA   <DK_SEC
  CMPA  #$10
  BHI   RECFI7
  LBSR  OPTABL
  LDA   #$03          erreur I/O
  BRA   RECFI1
RECFI7
  CLRB  
  BRA   RECFI9
RECFI8
  LDB   $0b,Y
  CMPB  <DKWEB
  BNE   RECFI7
  LDB   $0c,Y
  CMPB  <DKWEC
  BNE   RECFI7
  LDB   <DK_SEC
  LDA   $0d,Y
  STA   <DKWF6
  CLR   <DKWF5
  LDX   $0e,Y
  STX   <DKWF7
  STY   <DKWFA
RECFI9
  STB   <DKWF9     numero de bloc alloue
  LBRA  RETOK

*------------------------------------------------------
* ALLOD = allocation d'un fichier
*------------------------------------------------------
ALLOD
  LDY   <DKWED     adresse de la FAT
  BSR   ALLOB4     test place libre
ALLOD1
  BLO   RECFI1     erreur plus de place
  STB   <DKWF6     numero de bloc
  LBSR  LDIR0      charge debut directory
ALLOD2  
  BLO   ALLOD1     sortie en erreur
  LDY   <DKWE9     pointeur sur buffer
  LDX   #$0008     compteur de noms dans X
ALLOD3
  LDB   ,Y         premier octet du nom 
  BEQ   ALLOD6     0=fichier efface
  LDA   #$05
  CMPB  #$ff
  BEQ   ALLOD6
  LEAY  $20,Y
  LEAX  -$01,X
  BNE   ALLOD3
  INC   <DK_SEC
  LDA   <DK_SEC
  CMPA  #$10
  BHI   ALLOD4
  LBSR  OPTABL     charge le secteur suivant 
  LDA   #$03
  BRA   ALLOD2
ALLOD4
  LDA   #$05       code "disque plein"
ALLOD5
  BRA   ALLOD1     sort avec erreur
ALLOD6
  LDX   <DKWE7
  LDB   <DKWF0     code operation logique
  CMPB  #$03
  BNE   ALLOD7
  LEAX  SCRDOS,PCR adresse nom de fichier
ALLOD7
  LDB   #$0a
ALLOD8
  LDA   B,X        recopie le nom de fichier
  STA   B,Y        dans le catalogue
  DECB  
  BGE   ALLOD8
  LDA   <DKWEB     type de fichier 
  STA   $0b,Y
  LDA   <DKWEC     recopie le numero de bloc
  LDB   <DKWF6     dans le catalogue
  STD   $0c,Y
  LBRA  ECRSE      ecriture du repertoire

*------------------------------------------------------
* ALLOB = allocation d'un bloc
*------------------------------------------------------
ALLOB
  LDB   <DKWF6     numero de bloc
  CMPB  #$50
  BHI   ALLOB3
ALLOB1
  TSTB  
ALLOB2
  BEQ   ALLOB4
  LDA   B,Y
  CMPA  #$ff
  BEQ   ALLOB8
  DECB  
  CMPB  #$50
  BLS   ALLOB1
ALLOB3
  ADDB  #$02
  CMPB  #$A1
  BRA   ALLOB2
ALLOB4
  LDB   #$50  
ALLOB5
  LDA   #$05
  CMPB  #$A0
  LBHI  RETERR
  LDA   B,Y
  CMPA  #$ff
  BEQ   ALLOB8
  PSHS  B
  SUBB  #$50
  NEGB  
  ADDB  #$50
  LDA   B,Y
  CMPA  #$ff
  BEQ   ALLOB6
  PULS  B  
  INCB  
  BRA   ALLOB5
ALLOB6
  LEAS  $01,S 
ALLOB8
  CLR   B,Y
  DECB  
  STB   <DKWF9     numero de bloc alloue
ALLOB9
  LBRA  RETOK

*------------------------------------------------------
* RECUP = liberation d'un bloc
*------------------------------------------------------
RECUP
  LDA   $0d,Y
  STA   <DKWF6     numero de bloc
  CLR   ,Y
  LBSR  ECRSE
  BLO   ALLOD5
  LDY   <DKWED
  LDB   <DKWF6     numero de bloc
RECUP1
  INCB  
  LDA   B,Y
  CLR   B,Y
  DEC   B,Y
  TFR   A,B
  CMPA  #$c0
  BLO   RECUP1
  BRA   ALLOB9

*------------------------------------------------------
* MAJCL = Mise a jour cluster
*------------------------------------------------------
MAJCL
  LDB   <DKWF6       numero bloc                              
  CLRA  
  LSRB               divise par 2
  STD   <DKWFB       numero piste
  INCA  
  STA   <DKWF5
  BHS   MAJCL9 
  LDA   #$09   
MAJCL9
  STA   <DKWFA 
  RTS   

*------------------------------------------------------
* COMMANDES CARTE SD
*------------------------------------------------------
CMD0
  FCB   $40           go iddle state
  FDB   $0000
  FDB   $0000
  FCB   $95           checksum obligatoire           
  FCB   $01           code retour attendu  
*------------------------------------------------------
CMD8
  FCB   $48           send interface condition
  FDB   $0000
  FDB   $01AA
  FCB   $87           checksum obligatoire  
  FCB   $00           code retour attendu  
*------------------------------------------------------
CMD55
  FCB   $77           application command
  FDB   $0000
  FDB   $0000
  FCB   $FF           checksum non testee  
  FCB   $00           code retour attendu  
*------------------------------------------------------
AC41
  FCB   $69           activate card initialization 
  FDB   $4000
  FDB   $0000 
  FCB   $FF           checksum non testee  
  FCB   $00           code retour attendu  
*------------------------------------------------------
CMD58
  FCB   $7A           read OCR
  FDB   $0000
  FDB   $0000
  FCB   $FF           checksum non testee
  FCB   $00           code retour attendu  
*------------------------------------------------------
 
*------------------------------------------------------
* Signature
*------------------------------------------------------
  ORG   $E6F8
  FCC   "CS91280"
 
*------------------------------------------------------
* Numero de banque de l'EPROM
*------------------------------------------------------
 IF &BANK=0
  FCB   $B0           signature SDMOTO  
 ELSE
  FCB   $B1           signature SDMO  
 ENDIF 
 
*------------------------------------------------------
* Commentaires
*------------------------------------------------------
  FCC   'Controleur de ca'
  FCC   'rte SD pour les '
  FCC   'ordinateurs Thom'
  FCC   'son TO et MO - S'
  FCC   'imule quatre uni'
  FCC   'tes de disquette'
  FCC   's 3"1/2 double d'
  FCC   'ensite - Banque '
  FCC   '0 pour interface'
  FCC   ' SDMOTO, banque '
  FCC   '1 pour interface'
  FCC   ' SDMO.          '
  FCC   '================'
  FCC   'Daniel Coulom   '
  FCC   'dcmoto.free.fr  '
  FCC   'Vers. 2016.09.02'
 
  END